#include <caml/mlvalues.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include "migrate.h"

extern int threads_alive;

void safe_read(int fd, void* buf, int size)
{
  int to_read=size;
  int readed;
  fd_set fds;

  FD_ZERO(&fds);
  FD_SET(fd,&fds);

  while(to_read > 0){
    readed=read(fd, &Byte(buf,0), to_read);
    if(readed <= 0){
      if(readed != EAGAIN){
	fprintf(stderr, "One of the computers crashed !\n");
	exit(1);
      }
      select(fd+1,&fds,NULL,NULL,NULL);
    }
    else to_read-=readed;
  }
}

void safe_write(int fd, void* buf, int size)
{
  int to_write=size;
  int written;
  fd_set fds;

  FD_ZERO(&fds);
  FD_SET(fd,&fds);

  while(to_write > 0){
    written=write(fd, &Byte(buf,0), to_write);
    if(written <= 0){
      select(fd+1,NULL,&fds,NULL,NULL);
    }
    else to_write-=written;
  }
}


value safe_ext_read(value fd_p,value buf_p, value size_p)
{
  safe_read(Int_val(fd_p), (void*)buf_p, Int_val(size_p));
}

value safe_ext_write(value fd_p, value buf_p, value size_p)
{
  safe_write(Int_val(fd_p), (void*) buf_p, Int_val(size_p));
}


value write_size(value out_chan,value size)
{
  int true_size=Int_val(size);
  safe_write(Int_val(out_chan), &true_size, sizeof(int));
}

value read_size(value in_chan)
{
  int size;
  safe_read(Int_val(in_chan), &size, sizeof(int));
  return Val_int(size);
}
