#ifndef MIGRATE_H
#define MIGRATE_H

#include <setjmp.h>
#include <caml/fail.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/callback.h>
#include <caml/mlvalues.h>


#define STACK_SIZE ((64 * 1024)-364)
#define MAX_FTHREAD 32
#define MAX_COMP 8
#define MAX_STATIC_FTHREAD (MAX_FTHREAD*MAX_COMP)


#define Before_switch(curr) \
  curr->bottom_of_stack = caml_bottom_of_stack; \
  curr->last_return_address = caml_last_return_address;\
  curr->gc_regs = caml_gc_regs; \
  curr->exception_pointer = caml_exception_pointer; \
  curr->local_roots = local_roots


#define After_switch(curr) \
  caml_bottom_of_stack = curr->bottom_of_stack; \
  caml_last_return_address = curr->last_return_address; \
  caml_gc_regs = curr->gc_regs; \
  caml_exception_pointer = curr->exception_pointer;\
  local_roots = curr->local_roots

typedef struct sched_st *sched_t;
typedef struct fthread_st *fthread_t;

struct sched_st{
  sched_t next;
  sched_t prev;
  int alive_ft;
  int status;
  fthread_t current;
};

struct fthread_st{
  int home;
  value arg_pass;
  fthread_t next;  // structure circulaire doublement chaine 
  fthread_t prev;
  int id;          // id du Fair Thread courant
  sched_t sched;   // scheduler auquel il est rattach
  int towait;      // nombre de tour en attente (0 si actif)
  value event_id;  // identificateur de l'vnement
  char *bottom_of_stack; 
  unsigned long last_return_address; 
  value *gc_regs; 
  char *exception_pointer; 
  jmp_buf exit_buf;
  struct caml__roots_block *local_roots; 
  jmp_buf context; // dernier contexte d'xcution
  char stack[STACK_SIZE];
};

extern char *caml_bottom_of_stack;
extern unsigned long caml_last_return_address;
extern value *caml_gc_regs;
extern char *caml_exception_pointer;
extern struct caml__roots_block *local_roots;

#endif
