type 'a t

type sched_t=Fthread.sched_t

external see : 'a -> unit = "see"
external get_scheduler : unit -> sched_t = "get_scheduler"

external register_demux : Fthread.t -> unit = "register_demux"
external get_roots_tuple : unit -> 'a= "get_roots_tuple"
external assign_roots : 'a -> unit= "assign_roots"

external marshal_curr_cont : Unix.file_descr -> int = "marshal_curr_cont"
external app_cont_from_chan : Unix.file_descr -> 'a -> unit = "apply_cont_from_chan"
external get_tuple : unit -> 'a = "get_tuple"
external do_nothing : 'a -> unit = "do_nothing"
external safe_read : Unix.file_descr -> String.t -> int -> unit = "safe_ext_read"
external safe_write : Unix.file_descr -> String.t -> int -> unit = "safe_ext_write"
external net_read : Unix.file_descr -> String.t -> int -> unit = "ext_net_read"
external net_write : Unix.file_descr -> String.t -> int -> unit = "ext_net_write"
external read_size : Unix.file_descr -> int = "read_size"
external write_size : Unix.file_descr -> int -> unit = "write_size"
external kill : (unit -> unit) -> unit = "kill"
external get_max_nbr : unit -> int = "max_nbr"
external unlink : unit -> unit = "unlink"
external link : Fthread.sched_t -> unit = "link"
external get_home : unit -> int = "get_home"
external reset_fthread : unit -> unit = "reset_fthread"
external threads_alive : unit -> int = "get_threads_alive"
external decr_threads_alive : unit -> unit = "decr_threads_alive"
external incr_threads_alive : unit -> unit = "incr_threads_alive"
external swap_with : int -> unit = "swap_with"
external swap_to_demux : unit -> unit = "swap_to_demux"
external get_fthread_mode : unit -> int = "get_fthread_mode"
external get_max_comp : unit -> int = "get_max_comp"

external low_static_fthread : (unit -> unit) -> int -> unit = "create_static_fthread"
external create_slot : (unit -> unit) -> int -> unit = "create_slot"

exception Too_many_computers
exception Migration_forbidden
exception Dynamic_migrate

type migrate_t = (int -> unit) * int

let computer=ref 0

let max_nbr=get_max_nbr();;

let soc_tab=Array.make max_nbr Unix.stdin;;
let sock_port=Array.make max_nbr 0;;
let sock_addr=Array.make max_nbr "";;
let listen_tab=Array.make max_nbr Unix.stdin;;

let nbr_comp=ref 0;;

exception Computer_list_missing;;
exception Parsing_error;;

let set_file s=
  let in_chan=open_in s in
  try
    while true do
      Scanf.fscanf in_chan "%d.%d.%d.%d:%d"
	(fun ad1 ad2 ad3 ad4 port ->
	let addr=string_of_int ad1 ^ "." ^ string_of_int ad2 ^ "." ^ string_of_int ad3 ^
	  "." ^ string_of_int ad4
	in
	sock_addr.(!nbr_comp) <- addr;
	sock_port.(!nbr_comp) <- port;
	incr nbr_comp);
    done
  with
    End_of_file -> close_in in_chan
  | _ -> raise Parsing_error
;;

let set_comp_nbr x=
  if x >= get_max_comp() then
    raise Too_many_computers;
  computer:=x;;

Arg.parse [("-mynum", Arg.Int(set_comp_nbr), "this computer number");
	   ("-list", Arg.String(set_file), "list of computer (file)")]
  (fun x -> ()) "";;

if !nbr_comp = 0 then
  raise Computer_list_missing;;

for i=0 to (!computer)-1 do
  let soc=Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  Unix.connect soc
    (Unix.ADDR_INET(Unix.inet_addr_of_string sock_addr.(i),sock_port.(i)));
  soc_tab.(i) <- soc
done;;

let last_comp= !nbr_comp-1 in
if !computer < last_comp then
  let soc=Unix.socket Unix.PF_INET Unix.SOCK_STREAM 0 in
  Unix.bind soc (Unix.ADDR_INET
		   (Unix.inet_addr_of_string sock_addr.(!computer),
		    sock_port.(!computer)));
  Unix.listen soc last_comp;
  for i=(!computer)+1 to last_comp do
    match Unix.accept soc with
      (fd, _) ->
	soc_tab.(i) <- fd
  done;;

let last_comp=ref (!nbr_comp-1);;

let get_migrate fd=
  listen_tab.(get_home()) <- Unix.stdin;
  let size=read_size fd in
  let s=String.make size '0' in
  ignore (safe_read fd s size);
  let tuple=Marshal.from_string s 0 in (* ligne modifie *)
  app_cont_from_chan fd tuple


let migrate index=
  if index != !computer then
    begin
      write_size soc_tab.(index) (get_home()); (* ATTENTION *)
      let tuple=get_roots_tuple() in
      let s=Marshal.to_string tuple [Marshal.Closures] in
      write_size soc_tab.(index) (String.length s);
      ignore (safe_write soc_tab.(index) s (String.length s));
      if marshal_curr_cont soc_tab.(index) = 1 then
	begin
	  assign_roots (get_tuple());
	  reset_fthread();
	  swap_to_demux();
	  do_nothing index;
	end
      else kill (fun () -> ());
      do_nothing index
    end
	
(*  assign_roots (Marshal.from_channel f_in);*)

let create f comp=
  if comp = !computer then
    low_static_fthread 
      (fun () ->
	Gc.minor();
	try
	  f comp;   (* ZONE SENSIBLE NE RIEN RAJOUTER *)
	with _ -> kill 
	    (fun () -> 
	      output_string stderr "FATAL ERROR Thread uncaught exception\n";
	      Pervasives.exit(1);
	      flush stderr))
      comp
;;

let fd_list=ref [];;

register_demux
    (Fthread.create Fthread.main_scheduler
       (fun x ->
	 for i=0 to !last_comp do
	   if i != !computer then
	     fd_list:=soc_tab.(i) :: !fd_list;
	 done;
	 let rec listen()=
	   let (ready_l,_,_)=Unix.select !fd_list [] [] 
	       (if threads_alive() = 1 then (-1.0) else 0.0) in
	   List.iter 
	     (fun el ->
	       if el != Unix.stdin then
		 let readed=read_size el in
		 listen_tab.(readed) <- el;
		 ignore(create_slot 
			  (fun () -> 
			    get_migrate el)
			  readed);
		 swap_with(readed);
	     ) ready_l;
	   Fthread.cooperate();
	   listen()
	 in
	 listen()) ())
;;

let cooperate=Fthread.cooperate;;

let create_scheduler=Fthread.create_scheduler;;

let suspend_scheduler=Fthread.suspend_scheduler;;

let resume_scheduler=Fthread.resume_scheduler;;

let signal=Fthread.signal;;

let awaitn=Fthread.awaitn;;

let link=Fthread.link;;

let exit=Fthread.exit
;;

let rec remove_list a l=
  match !l with
    [] -> ()
  | x :: rl ->
      if x = a then
	l:=rl
      else remove_list a (ref rl)
;;

let read_line=
  let rec loop alive=
    let (ready,_,_)=Unix.select [Unix.stdin] [] [] (0.0) in
    if ready=[] then 
      begin 
	if alive then
	  begin
	    fd_list:=Unix.stdin :: !fd_list;
	    decr_threads_alive();
	  end;
	Fthread.cooperate();
	loop false
      end
    else
      begin
	incr_threads_alive();
	remove_list Unix.stdin fd_list;
	Pervasives.read_line()
      end
  in
  fun () -> loop true
;;

let read fd buf off len=
  let rec loop alive=
    let (ready,_,_)=Unix.select [Unix.stdin] [] [] (0.0) in
    if ready=[] then 
      begin 
	if alive then
	  begin
	    fd_list:=fd :: !fd_list;
	    decr_threads_alive();
	  end;
	Fthread.cooperate();
	loop false
      end
    else
      begin
	incr_threads_alive();
	remove_list fd fd_list;
	Unix.read fd buf off len
      end
  in
  fun () -> loop true
;;
